# tessif/frused/themes.py
# -*- coding: utf-8 -*-
"""
:mod:`~tessif.frused.themes` is a :mod:`tessif` subpackage grouping color and
hatch themes to :`tessif.frused.namedtuples.NodeColorGroupings` for convenient
and automated access.
"""
import collections
from itertools import cycle
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
#
from tessif.frused.namedtuples import NodeColorGroupings, Uid
import tessif.frused.spellings as spellings


colors = NodeColorGroupings(
    component=collections.OrderedDict([
        ('bus', '#9999ff'),  # websafe very light blue
        ('sink', '#9999ff'),  # websafe very light blue
        ('storage', '#9999ff'),  # websafe very light blue
        ('source', '#9999ff'),  # websafe very light blue
        ('transformer', '#9999ff'),  # websafe very light blue
        ('connector', '#9999ff'),  # websafe very light blue
    ]),
    name=collections.OrderedDict([
        ('renewables', '#00ff00'),  # websafe mostly pure lime green
        ('photovoltaic', '#ff9900'),  # websafe mostly pure orange
        ('solarthermal', '#ff0099'),  # websafe mostly pure pink
        ('onshore', '#99ccff'),  # websafe very light blue
        ('offshore', '#00ccff'),  # websafe mosty pure cyan
        ('hydro_electric', '#000099'),  # websafe dark blue
        ('combined_heat_power', '#9933cc'),  # websafe strong violet
        ('power_plant', '#ff6600'),  # websafe mostly pure orange
        ('heat_plant', '#b30000'),  # strong red
        ('electrical_line', '#ffcc00'),  # websafe mostly pure yellow
        ('gas_station', '#6633cc'),
        ('gas_pipeline', '#336666'),  # websafe very dark desaturated cyan
        ('gas_delivery', '#006666'),  # websafe very dark cyan
        ('oil_pipeline', '#666666'),  # websafe very dark grey
        ('oil_delivery', '#333333'),  # websafe even verier dark grey
        ('hydro_electrical_storage', '#0000cc'),  # websafe strong blue
        ('electro_chemical_storage', '#ccff00'),  # websafe mostly pure yellow
        ('electro_mechanical_storage', '#999900'),  # websafe dark yellow
        ('thermal_energy_storage', '#cc0033'),  # websafe strong red
        ('power2x', '#669999'),  # websafe mostly desaturated dark cyan
        ('power2heat', '#b30000'),  # strong red
        ('imported', '#ff6600'),  # websafe orange
        ('backup', '#990099'),  # websafe dark magenta
        ('demand', '#330099'),  # websafe dark violet
        ('export', '#006600'),  # websafe dark green
        ('excess', '#cc0000'),  # websafe pure red
        ('connector', '#669999'),  # websafe mostly desaturated dark cyan
    ]),
    carrier=collections.OrderedDict([
        ('solar', '#ff9900'),  # websafe mostly pure orange
        ('wind', '#00ccff'),  # websafe mosty pure cyan
        ('water', '#000099'),  # websafe dark blue
        ('biomass', '#009900'),  # websafe dark lime green
        ('gas', '#336666'),  # websafe very dark desaturated cyan
        ('oil', '#666666'),  # websafe very dark gray
        ('lignite', '#993300'),  # websafe dark orange
        ('hardcoal', '#000000'),  # websafe black
        ('nuclear', '#cccc00'),  # websafe strong yellow
        ('electricity', '#FFD700'),  # websafe pure yellow
        ('steam', '#cc0033'),  # websafe crimson
        ('hot_water', '#ff3300'),  # websafe mostly pure red
    ]),
    sector=collections.OrderedDict([
        ('power', '#ffff33'),  # websafe vivid yellow
        ('heat', '#FF0000'),  # websafe red
        ('mobility', '#669999'),  # websafe mostly desaturated dard cyan
        ('coupled', '#6633cc'),  # websafe strong violet
    ]),)
""" Tessif color themes.  Stored inside a
:attr:`~tessif.frused.namedtuples.NodeColorGroupings`
:class:`~typing.NamedTuple`. Assembled using `colorhexa
<https://www.colorhexa.com/>`_
Categorized by :attr:`~tessif.frused.namedtuples.NodeColorGroupings`

For each category there is a mapping of identifiers to colors

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    import tessif.frused.themes as themes
    import pandas
    import dutils

    for pos, dct in enumerate(themes.colors):
        category = themes.colors._fields[pos]

        path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_colors.png'.format(category))
        fig = themes._plot_colortable(
            dct, str(category),
            sort_colors=False, emptycols=0)
        fig.savefig(path)


        csv_path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}.csv'.format(category))

        df = dutils.to_dataframe(
            dct, columns=2, fillvalue='')

        df.to_csv(csv_path, header=None, index=None, na_rep='None')

|

**Component Grouped Colors:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/component.csv

.. image:: themes/component_colors.png
    :align: center
    :alt: Image showing the supported component colors

|

**Named Grouped Colors:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/name.csv
    :stub-columns: 1

.. image:: themes/name_colors.png
    :align: center
    :alt: Image showing the supported name colors

|

**Carrier Grouped Colors:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/carrier.csv
    :stub-columns: 1

.. image:: themes/carrier_colors.png
    :align: center
    :alt: Image showing the supported carrier colors

|

**Sector Grouped Colors:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/sector.csv
    :stub-columns: 1

.. image:: themes/sector_colors.png
    :align: center
    :alt: Image showing the supported sector colors
"""


cmaps = NodeColorGroupings(
    component=collections.OrderedDict([
        ('bus', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                 '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
        ('sink', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                  '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
        ('storage', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                     '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
        ('source', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                    '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
        ('transformer', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                         '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
        ('connector', ['#4d4dff', '#6666ff', '#8080ff', '#9999ff',
                       '#b3b3ff', '#ccccff', '#e6e6ff']),  # very light blue
    ]),
    carrier=collections.OrderedDict([
        ('solar', ['#ff5900', '#ff6f00', '#ff8400',
                   '#ff9900', '#ffae00', '#ffc400', '#ffd900']),  # orange
        ('wind', ['#002aff', '#008cff', '#00b7ff', '#00ccff',
                  '#00e1ff', '#00f7ff', '#00fff2']),  # cyan
        ('water', ['#00004d', '#190099', '#0d0099',
                   '#000099', '#000d99', '#001a99', '#002699']),  # dark blue
        ('biomass', ['#004d00', '#006600', '#008000', '#009900',
                     '#00b300', '#00cc00', '#00e600']),  # dark lime
        ('gas', ['#1a3333', '#224444', '#2b5555', '#336666',
                 '#3c7777', '#448888', '#4d9999']),  # dark desaturated cyan
        ('oil', ['#404040', '#4d4d4d', '#595959', '#666666',
                 '#737373', '#808080', '#8c8c8c']),  # very dark grey
        ('lignite', ['#4d1a00', '#662200', '#802b00', '#993300',
                     '#b33c00', '#cc4400', '#e64d00']),  # dark brown
        ('hardcoal', ['#000000', '#0d0d0d', '#191919', '#262626',
                      '#333333', '#404040', '#4c4c4c']),  # black
        ('nuclear', ['#808000', '#999900', '#b3b300', '#cccc00',
                     '#e6e600', '#ffff00', '#ffff1a']),  # strong yellow
        ('electricity', ['#b39700', '#ccac00', '#e6c200', '#ffd700',
                         '#ffdb1a', '#ffdf33', '#ffe34d']),  # pure yellow
        ('steam', ['#800020', '#990026', '#b3002a', '#cc0033',
                   '#e60039', '#ff0040', '#ff1a53']),  # crimson
        ('hot_water', ['#b32400', '#cc2900', '#e62e00', '#ff3300',
                       '#ff471a', '#ff5c33', '#ff704d']),  # pure red
    ]),
    sector=collections.OrderedDict([
        ('power', ['#ff5900', '#ff6f00', '#ff8400', '#ff9900',
                   '#ffae00', '#ffc400', '#ffd900']),  # vivid yellow
        ('heat', ['#b30000', '#cc0000', '#e60000',
                  '#ff0000', '#ff1a1a', '#ff3333', '#ff4d4d']),  # red
        ('mobility', ['#476b6b', '#527a7a', '#5c8a8a', '#669999', '#75a3a3',
                      '#85adad', '#94b8b8']),  # desaturated dark cyan
        ('coupled', ['#47248f', '#5229a3', '#5c2eb8', '#6633cc',
                     '#7547d1', '#855cd6', '#9470db', ]),  # strong violet
    ]),
    name=collections.OrderedDict([
        ('renewables', ['#00b300', '#00cc00', '#00e600', '#00ff00',
                        '#1aff1a', '#33ff33', '#4dff4d']),  # lime green
        ('photovoltaic', ['#b37400', '#cc8400', '#e69500', '#ffa500',
                          '#ffae1a', '#ffb733', '#ffc04d']),  # pure orange
        ('solarthermal', ['#b300b6', '#cc007a', '#e6008a', '#ff0099',
                          '#ff1aa3', '#ff33ad', '#ff4db8']),  # pure pink
        ('offshore', ['#4da6ff', '#66b3ff', '#80bfff', '#99ccff',
                      '#b3d9ff', '#cce6ff', '#e6f2ff']),  # pure cyan
        ('onshore', ['#002aff', '#008cff', '#00b7ff', '#00ccff',
                     '#00e1ff', '#00f7ff', '#00fff2']),  # very light blue
        ('hydro_electric', ['#00004d', '#190099', '#0d0099',
                            '#000099', '#000d99', '#001a99', '#002699']),  # dark blue
        ('combined_heat_power', ['#6b248f', '#7a29a3', '#8a2eb8', '#9933cc',
                                 '#a347d1', '#ad5cd6', '#b870db']),  # violet
        ('power_plant', ['#b34700', '#cc5200', '#c65c00', '#ff6600',
                         '#ff751a', '#ff8533', '#ff944d']),  # pure orange
        ('heat_plant', ['#670000', '#800000', '#9a0000', '#b30000',
                        '#cd0000', '#e60000', '#ff0000']),  # strong red
        ('electrical_line', ['#b38f00', '#cca300', '#e6b800', '#ffcc00',
                             '#ffd11a', '#ffd633', '#ffdb4d']),  # pure yellow
        ('gas_pipeline', ['#1a3333', '#224444', '#2b5555', '#336666',
                          '#3c7777', '#448888', '#4d9999']),  # v. dark d. cyan
        ('gas_delivery', ['#003434', '#004d4d', '#006767', '#008080',
                          '#009a9a', '#00b3b3', '#00cdcd']),  # very dark cyan
        ('oil_pipeline', ['#404040', '#4d4d4d', '#595959', '#666666',
                          '#737373', '#808080', '#8c8c8c']),  # very dark grey
        ('oil_delivery', ['#0d0d0d', '#1a1a1a', '#262626', '#333333',
                          '#404040', '#4d4d4d', '#595959']),  # darker grey
        ('hydro_electrical_storage', ['#000080', '#000099', '#0000b3',
                                      '#0000cc', '#0000c6',
                                      '#0000ff', '#1a1aff']),  # strong blue
        ('electro_chemical_storage', ['#8fb300', '#a3cc00', '#b8e600',
                                      '#ccff00', '#d1ff1a', '#d6ff33',
                                      '#dbff4d']),  # pure yellow
        ('electro_mechanical_storage', ['#343400', '#4d4d00', '#676700',
                                        '#808000', '#9a9a00', '#b3b300',
                                        '#cdcd00']),  # dark yellow
        ('thermal_energy_storage', ['#800020', '#990026', '#b3002a',
                                    '#cc0033', '#e60039', '#ff0040',
                                    '#ff1a53']),  # strong red
        ('power2x', ['#476b6b', '#527a7a', '#5c8a8a', '#669999',
                     '#75a3a3', '#85adad', '#94b8b8']),  # desaturat. dark cyan
        ('power2heat', ['#670000', '#800000', '#9a0000', '#b30000',
                        '#cd0000', '#e60000', '#ff0000']),  # strong red
        ('imported', ['#b33500', '#cc3d00', '#e64400', '#ff4c00',
                      '#ff5e1a', '#ff7033', '#ff824d']),  # orange
        ('backup', ['#340034', '#4d004d', '#670067', '#800080',
                    '#9a009a', '#b300b3', '#cd00cd']),  # dark magenta
        ('demand', ['#1a004d', '#220066', '#2b0080', '#330099',
                    '#3b00b3', '#4400cc', '#4c00e6']),  # dark violet
        ('export', ['#001800', '#003100', '#004b00', '#006400',
                    '#007e00', '#009700', '#00b100']),  # dark green
        ('excess', ['#800000', '#990000', '#b30000', '#cc0000',
                    '#e60000', '#ff0000', '#ff1a1a']),  # pure red
    ]),
)
""" Tessif colormap themes. Stored inside a
:attr:`~tessif.frused.namedtuples.NodeColorGroupings`
:class:`~typing.NamedTuple`. Assembled using `colorhexa
<https://www.colorhexa.com/>`_
Categorized by :attr:`~tessif.frused.namedtuples.NodeColorGroupings`

For each category there is a mapping of identifiers to colors

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    import tessif.frused.themes as themes
    import pandas

    for pos, dct in enumerate(themes.cmaps):
        category = themes.cmaps._fields[pos]

        path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_colormaps.png'.format(category))
        fig = themes._plot_color_gradients(category, dct)
        fig.savefig(path)

        csv_path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_colormaps.csv'.format(category))

        df = pandas.DataFrame(data=dct.values(),
                              index=dct.keys())

        df.to_csv(csv_path, header=None)

|

**Component Grouped Colormaps:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/component_colormaps.csv

.. image:: themes/component_colormaps.png
    :align: center
    :alt: Image showing the supported component colormaps

|

**Name Grouped Colormaps:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/name_colormaps.csv
    :stub-columns: 1

.. image:: themes/name_colormaps.png
    :align: center
    :alt: Image showing the supported name colormaps

|

**Carrier Grouped Colormaps:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/carrier_colormaps.csv
    :stub-columns: 1

.. image:: themes/carrier_colormaps.png
    :align: center
    :alt: Image showing the supported carrier colormaps

|

**Sector Grouped Colormaps:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/sector_colormaps.csv
    :stub-columns: 1

.. image:: themes/sector_colormaps.png
    :align: center
    :alt: Image showing the supported sector colormap
"""

ccycles = NodeColorGroupings(*[
    collections.OrderedDict([
        (catgry, cycle(color_list))
        for catgry, color_list in cmaps._asdict()[key].items()])
    for key in cmaps._asdict().keys()])
"""
:mod:`~tessif` cycled colormaps. Stored inside a
:attr:`~tessif.frused.namedtuples.NodeColorGroupings`
:class:`~typing.NamedTuple`.
"""

hatches = NodeColorGroupings(
    component=collections.OrderedDict([
        ('bus', '/'),
        ('sink', '\\'),
        ('source', '.'),
        ('storage', '|'),
        ('transformer', '-'),
        ('connector', 'x'),
    ]),
    carrier=collections.OrderedDict([
        ('solar', '/'),
        ('wind', '\\'),
        ('water', '.'),
        ('biomass', '|'),
        ('gas', '-'),
        ('oil', '+/'),
        ('lignite', 'o'),
        ('hardcoal', '.'),
        ('nuclear', '*'),
        ('electricity', '//'),
        ('steam', '\\\\'),
        ('hot_water', '||'),
    ]),
    sector=collections.OrderedDict([
        ('power', '/'),
        ('heat', '\\'),
        ('mobility', '|'),
        ('coupled', 'x'),
    ]),
    name=collections.OrderedDict([
        ('photovoltaic', '/'),
        ('solarthermal', '//'),
        ('onshore', '\\'),
        ('offshore', '\\\\'),
        ('hydro_electric', 'o'),
        ('combined_heat_power', '+'),
        ('power_plant', '|'),
        ('heat_plant', '-'),
        ('electrical_line', '++'),
        ('gas_pipeline', '--'),
        ('gas_delivery', '---'),
        ('oil_pipeline', '+/'),
        ('oil_delivery', '+/+/'),
        ('bus', '.'),
        ('hydro_electrical_storage', 'oo'),
        ('electro_chemical_storage', 'xx'),
        ('electro_mechanical_storage', '||'),
        ('thermal_energy_storage', '--'),
        ('power2x', 'xxx'),
    ]),
)
"""
:mod:`~tessif` hatch themes. Stored inside a
:attr:`~tessif.frused.namedtuples.NodeColorGroupings`
:class:`~typing.NamedTuple`.

For each category there is a mapping of identifiers to hatches

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    import tessif.frused.themes as themes
    import pandas
    import dutils

    for pos, dct in enumerate(themes.hatches):
        category = themes.hatches._fields[pos]

        path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_hatches.png'.format(category))
        fig = themes._plot_hatches(category, dct)
        fig.savefig(path)

        csv_path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_hatches.csv'.format(category))

        for key, value in dct.copy().items():
            dct[key] = '{lit_quote}{value}{lit_quote}'.format(
                lit_quote='``', value=value)

        df = dutils.to_dataframe(dct, columns=2, fillvalue='')

        df.to_csv(csv_path, header=None)

|

**Component Grouped Hatches:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/component_hatches.csv
    :stub-columns: 1

.. image:: themes/component_hatches.png
    :align: center
    :alt: Image showing the supported name hatches

|

**Name Grouped Hatches:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/name_hatches.csv
    :stub-columns: 1

.. image:: themes/name_hatches.png
    :align: center
    :alt: Image showing the supported name hatches

|

**Carrier Grouped Hatches:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/carrier_hatches.csv
    :stub-columns: 1

.. image:: themes/carrier_hatches.png
    :align: center
    :alt: Image showing the supported carrier hatches

|

**Sector Grouped Hatches:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/sector_hatches.csv
    :stub-columns: 1

.. image:: themes/sector_hatches.png
    :align: center
    :alt: Image showing the supported sector hatches
"""


hmaps = {
    'sector': collections.OrderedDict([
        ('power', ['/', '\\', '//', '\\\\', '///', '\\\\\\']),
        ('heat', ['-', '|', '--', '||', '---', '|||']),
        ('mobility', ['x', '+/', 'xx', '+/+/', 'xxx', '+/+/+/']),
    ]),
}
"""
Mapping of :mod:`~tessif` hatchmaps. Usefull when plotting sector grouped
results to distinguish the individual components without coloring.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    import tessif.frused.themes as themes
    import pandas
    import dutils

    for category, dct in themes.hmaps.items():

        csv_path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'themes', '{}_hatchmaps.csv'.format(category))

        for key, value in dct.copy().items():
            new_list = []
            for i in value:
                new_list.append(
                    '{lit_quote}{i}{lit_quote}'.format(lit_quote='``', i=i))
            dct[key] = new_list

        df = pandas.DataFrame(data=dct.values(),
                              index=dct.keys())

        df.to_csv(csv_path, header=None)

|

**Sector Grouped Hatchmap:**

.. csv-table:: internal representation
    :file: source/api/frused/themes/sector_hatchmaps.csv
    :stub-columns: 1
"""


def _plot_colortable(colors, title, sort_colors=True, emptycols=0):
    """
    Utility for plotting color themes. Inspired by
    matplotlib's `documentation
    <https://matplotlib.org/3.2.2/gallery/color/named_colors.html>`_.
    """

    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12
    topmargin = 40

    # Sort colors by hue, saturation, value and name.
    if sort_colors is True:
        by_hsv = sorted((tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(color))),
                         name)
                        for name, color in colors.items())
        names = [name for hsv, name in by_hsv]
    else:
        names = list(colors)

    n = len(names)
    ncols = 4 - emptycols
    nrows = n // ncols + int(n % ncols > 0)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + margin + topmargin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-topmargin)/height)
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()
    ax.set_title(title, fontsize=24, loc="left", pad=10)

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        swatch_end_x = cell_width * col + swatch_width
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=14,
                horizontalalignment='left',
                verticalalignment='center')

        ax.hlines(y, swatch_start_x, swatch_end_x,
                  color=colors[name], linewidth=18)

    return fig


def _plot_color_gradients(category, cmap_dct):
    """
    Utility for plotting colormap examples. Inspired by
    matplotlib's `documentation
    <https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html>`_.
    """
    import numpy as np

    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))

    fig, axes = plt.subplots(nrows=len(cmap_dct))
    fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)
    axes[0].set_title(category + ' colormaps', fontsize=14)

    for ax, name in zip(axes, cmap_dct):
        ax.imshow(gradient, aspect='auto',
                  cmap=mcolors.ListedColormap(cmap_dct[name]))
        pos = list(ax.get_position().bounds)
        x_text = pos[0] - 0.01
        y_text = pos[1] + pos[3]/2.
        fig.text(x_text, y_text, name, va='center', ha='right', fontsize=10)

    # Turn off *all* ticks & spines, not just the ones with colormaps.
    for ax in axes:
        ax.set_axis_off()

    return fig


def _plot_hatches(category, hatch_map, sort_colors=True, emptycols=0):
    """
    Utility for plotting color themes. Inspired by
    matplotlib's `documentation
    <https://matplotlib.org/3.2.2/gallery/color/named_colors.html>`_.

    And a pach hatching example from `so
    <https://stackoverflow.com/a/25185432>`
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12
    topmargin = 40

    n = len(hatch_map)
    ncols = 4 - emptycols
    nrows = n // ncols + int(n % ncols > 0)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + margin + topmargin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-topmargin)/height)
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()
    ax.set_title(category, fontsize=24, loc="left", pad=10)

    for i, name in enumerate(hatch_map):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        swatch_end_x = cell_width * col + swatch_width
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=14,
                horizontalalignment='left',
                verticalalignment='center')

        ax.add_patch(patches.Rectangle(
            (swatch_start_x, y-9), swatch_end_x-swatch_start_x, 18,
            hatch=hatch_map[name], fill=False, snap=False, linewidth=0))

    return fig


def match_theme(strings, theme='colors', grouping='name',):
    """
    Match a collection of strings to tessf's theme groupings

    Parameters
    ---------
    strings: ~collections.abc.Iterable
        Iterable of strings which are tried to be matched to registerd theme
        values.
    theme: str
        String specifying one of :mod:`tessif's themes <tessif.frused.themes>`
        like ``colors``, etc.
    grouping: str
        String specifying one of the
        :attr:`~tessif.frused.namedtuples.NodeColorGroupings` the
        :paramref:`~match_theme.theme` provides

    Return
    ------
    dict
        Mapping of the provided strings to the matched theme values

    Example
    -------
    Create a :class:`tessif energy system
    <tessif.model.energy_system.AbstractEnergySystem>` and create a color
    mapping using it's nodes:

    1. Create the tessif energy system using the :ref:`Example Hub <Examples>`:

       >>> import tessif.examples.data.tsf.py_hard as coded_examples
       >>> tessif_es = coded_examples.create_fpwe()

    2. Create a list of node names:
       >>> node_names = [str(node.uid) for node in tessif_es.nodes]

    3. Match the node colors:

       >>> from tessif.frused.themes import match_theme
       >>> node_colors = match_theme(
       ...     strings=node_names, theme='colors', grouping='name')

       >>> import pprint
       >>> pprint.pprint(node_colors)
       {'Battery': '#ccff00',
        'Demand': '#330099',
        'Gas Station': '#6633cc',
        'Generator': '#ff6600',
        'Pipeline': '#336666',
        'Powerline': '#ffcc00',
        'Solar Panel': '#ff9900'}
    """
    matched_theme = dict()
    for string in strings:

        # get requested theme group of this module
        theme_group = getattr(globals().get(theme), grouping)

        # reconstruct the uid from its string representation:
        uid = Uid.reconstruct(string)

        # match reconstructed uid groupings attribute to spellings:
        for registered_key in theme_group:
            if any(tag == getattr(uid, grouping) for tag in getattr(
                    spellings, registered_key)):
                matched_theme[string] = theme_group[registered_key]

    return matched_theme



# src/tessif/model/energy_system.py
# pylint: disable=too-few-public-methods
"""Dummy energy_system module to enebale tessif-examples."""

# standard library
# import shutil
# import tempfile
# from pathlib import Path
import os
import pickle
import subprocess
import tempfile

# third pary packages
import pandas as pd

# local packages
import tessif.model.components as tessif_components
import tessif.frused.namedtuples as nts

uid_seperator = "_"

registered_plugins = {
    "oemof-4.4": "tessif-oemof-4.4",
    "omeof-latest": "tessif-oemof-4.4",
    "oemof": "tessif-oemof-4.4",
    "omf": "tessif-oemof-4.4",
}


class AbstractEnergySystem:
    """
    Aggregate tessif's abstract components into an energy system.

    Parameters
    ----------
    uid: ~collections.abc.Hashable
        Hashable unique identifier. Usually a string aka a name.
    busses: ~collections.abc.Iterable
        Iterable of :class:`~tessif.model.components.Bus`
        objects to be added to the energy system
    sinks: ~collections.abc.Iterable
        Iterable of :class:`~tessif.model.components.Sink`
        objects to be added to the energy system
    sources: ~collections.abc.Iterable
        Iterable of :class:`~tessif.model.components.Source`
        objects to be added to the energy system
    transformers: ~collections.abc.Iterable
        Iterable of
        :class:`~tessif.model.components.Transformer`
        objects to be added to the energy system
    storages: ~collections.abc.Iterable
        Iterable of
        :class:`~tessif.model.components.Storage`
        objects to be added to the energy system
    timeframe: pandas.DatetimeIndex
        Datetime index representing the evaluated timeframe. Explicitly
        stating:

            - initial datatime
              (0th element of the :class:`pandas.DatetimeIndex`)
            - number of timesteps (length of :class:`pandas.DatetimeIndex`)
            - temporal resolution (:attr:`pandas.DatetimeIndex.freq`)

        For example::

            idx = pd.DatetimeIndex(
                data=pd.date_range(
                    '2016-01-01 00:00:00', periods=11, freq='H'))

    global_constraints: dict, default={'emissions': float('+inf')}
        Dictionary of :class:`numeric <numbers.Number>` values mapped to
        global constraint naming :class:`strings <str>`.

        Recognized constraint keys are:

            - ``emissions``
            - ...

        For a more detailed explanation see the user guide's section:
        :ref:`Secondary_Objectives`
    """

    def __init__(self, uid, *args, **kwargs):

        self._uid = uid

        kwargs_and_defaults = {
            'busses': (),
            'chps': (),
            'connectors': (),
            'sinks': (),
            'sources': (),
            'transformers': (),
            'storages': (),
            'timeframe': pd.Series(),
            'global_constraints': {'emissions': float('+inf')}}

        for kwarg, default in kwargs_and_defaults.copy().items():

            # overwrite default if user provided key word argument:
            kwargs_and_defaults[kwarg] = kwargs.get(kwarg, default)

            # initialize instance respective instance attribute:
            if kwarg == 'global_constraints':
                setattr(self, '_{}'.format(kwarg), kwargs_and_defaults[kwarg])

            elif kwarg != 'timeframe':
                setattr(self, '_{}'.format(kwarg),
                        tuple(kwargs_and_defaults[kwarg]))
            else:
                self._timeframe = kwargs_and_defaults[kwarg]

        self._es_attributes = tuple(kwargs_and_defaults.keys())

    def __repr__(self):
        return '{!s}({!r})'.format(self.__class__, self.__dict__)

    def __str__(self):
        return '{!s}(\n'.format(self.__class__) + ',\n'.join([
            *['    {!r}={!r}'.format(
                k.lstrip('_'), v) for k, v in self.__dict__.items()],
            ')'
        ])

    @property
    def uid(self):
        """:class:`~collections.abc.Hashable` unique identifier. Usually a
        string aka a name.
        """
        return self._uid

    @property
    def busses(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Bus` objects part
        of the energy system.
        """
        for bus in self._busses:
            yield bus

    @property
    def chps(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.CHP`
        objects part of the energy system.
        """
        for chp in self._chps:
            yield chp

    @property
    def connectors(self):
        """ :class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Connectors` objects part
        of the energy system.
        """

        for connector in self._connectors:
            yield connector

    @property
    def sources(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Source` objects
        part of the energy system.
        """
        for source in self._sources:
            yield source

    @property
    def sinks(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Sink` objects part
        of the energy system.
        """
        for sink in self._sinks:
            yield sink

    @property
    def transformers(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Transformer`
        objects part of the energy system.
        """
        for transformer in self._transformers:
            yield transformer

    @property
    def storages(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.model.components.Storage` objects
        part of the energy system.
        """
        for storage in self._storages:
            yield storage

    @property
    def nodes(self):
        """
        :class:`~collections.abc.Generator` yielding this energy system's
        components.
        """

        component_types = ['busses', 'chps', 'sources',
                           'sinks', 'transformers', 'storages',
                           'connectors']

        for component_type in component_types:
            for component in getattr(self, component_type):
                yield component

    @property
    def edges(self):
        """:class:`~collections.abc.Generator` of
        :class:`~tessif.frused.namedtuples.Edge`
        :class:`NamedTuples<typing.Namedtuple>` representing graph like `edges
        <https://en.wikipedia.org/wiki/Glossary_of_graph_theory_terms#edge>`_.
        """
        # All edge information are stored inside the bus objects..
        for bus in self.busses:
            # Bus incoming edge should contain node.uid and bus.uid:
            for inflow in bus.inputs:
                # so find out node uid by string comparison with
                # every single node:
                for node in self.nodes:
                    if inflow.split('.')[0] == node.uid.name:
                        edge = nts.Edge(str(node.uid), str(bus.uid))
                        yield edge

            # Bus leaving edges should contain bus.uid and node.uid:
            for outflow in bus.outputs:
                # so find out node uid by string comparison with
                # every single node:
                for node in self.nodes:
                    if outflow.split('.')[0] == node.uid.name:
                        edge = nts.Edge(str(bus.uid), str(node.uid))
                        yield edge

        # ... except for the edges build by the connectors
        for connector in self.connectors:
            for inflow in connector.inputs:
                edge = nts.Edge(inflow, str(connector.uid))
                yield edge

            for outflow in connector.outputs:
                edge = nts.Edge(str(connector.uid), outflow)
                yield edge

    @property
    def global_constraints(self):
        """
        :class:`Dictionary <dict>` of :class:`numeric <numbers.Number>`
        values mapped to global constraint naming :class:`strings <str>`
        currently respected by the energy system.
        """
        return self._global_constraints

    def _edge_carriers(self):
        """
        Extract carrier information out of busses and connectors.
        """
        _ecarriers = {}
        # All edge information are stored inside the bus objects
        for bus in self.busses:
            # Bus incoming edge should contain node.uid and bus.uid:
            for inflow in bus.inputs:
                # so find out node uid by string comparison with
                # every single node:
                for node in self.nodes:
                    if inflow.split('.')[0] == node.uid.name:
                        edge = nts.Edge(str(node.uid), str(bus.uid))
                        _ecarriers[edge] = inflow.split('.')[1]

            # Bus leaving edges should contain bus.uid and node.uid:
            for outflow in bus.outputs:
                # so find out node uid by string comparison with
                # every single node:
                for node in self.nodes:
                    if outflow.split('.')[0] == node.uid.name:
                        edge = nts.Edge(str(bus.uid), str(node.uid))
                        _ecarriers[edge] = inflow.split('.')[1]

        for connector in self.connectors:
            for inflow in connector.inputs:
                for bus in self.busses:
                    if inflow == str(bus.uid):
                        edge = nts.Edge(inflow, str(connector.uid))
                        _ecarriers[edge] = list(bus.outputs)[0].split('.')[1]

            for outflow in connector.outputs:
                for bus in self.busses:
                    if outflow == str(bus.uid):
                        edge = nts.Edge(str(connector.uid), outflow)
                        _ecarriers[edge] = list(bus.inputs)[0].split('.')[1]

        return _ecarriers

    def connect(self, energy_system, connecting_busses, connection_uid):
        """Connect another :class:`AbstractEnergySystem` object to this one.

        Parameters
        ----------
        energy_system: tessif.model.energy_system.AbstractEnergySystem
            Energy system object to be connected to this energy system via the
            respective :paramref:`~connect_energy_systems.connecting_busses`.

            The :paramref:`energy_system's <connect.energy_system>`
            :class:`~tessif.model.components.Bus` specified in
            :paramref:`connecting_busses[1] <connect.connecting_busses>` will
            be connected to this energy system's
            :class:`~tessif.model.components.Bus` specified in
            :paramref:`connecting_busses[0] <connect.connecting_busses>`. So
            it's :attr:`~tessif.model.components.AbstractEsComponent.uid`
            must be found in this energy system.

        connecting_busses: ~collections.abc.tuple
            Tuple of :attr:`Uids <tessif.frused.namedtuples.Uid>` string
            representation specifying the busses with which the energy systems
            will be connected.

            The :paramref:`energy_system's <connect.energy_system>`
            :class:`~tessif.model.components.Bus` specified in
            :paramref:`connecting_busses[1] <connect.connecting_busses>` will
            be connected to this energy system's
            :class:`~tessif.model.components.Bus` specified in
            :paramref:`connecting_busses[0] <connect.connecting_busses>`.

        connection_uid: tessif.frused.namedtuples.Uid
            Uid of the :class:`~tessif.model.components.Connector` object
            created for connecting the energy system.


        Return
        ------
        tessif.model.energy_system.AbstractEnergySystem
            The energy system created by connecting the
            :paramref:`~connect.energy_system` to this energy system.
        """

        components_to_add = list()

        for component in energy_system.nodes:
            # is current component to be connected ?
            if not str(component.uid) == connecting_busses[1]:
                # no, so just prepare it for adding to the new es
                components_to_add.append(component)
            else:
                # yes it's to be connected, so create a connector:
                connector = tessif_components.Connector(
                    **connection_uid._asdict(),
                    interfaces=(connecting_busses[0],
                                connecting_busses[1]))

                components_to_add.append(connector)
                components_to_add.append(component)

        connected_es = self.from_components(
            uid=self.uid,
            timeframe=self.timeframe,
            global_constraints=self.global_constraints,
            components=set([*self.nodes, *components_to_add]))

        return connected_es

    def duplicate(self, prefix='', separator='_', suffix='copy'):
        """
        Duplicate the energy system and return it. Potentially modify
        the node names.

        Parameters
        ----------
        prefix: str, default=''
           String added to the beginning of every node's :attr:`Uid.name
           <tessif.frused.namedtuples.Uid>`, separated by
           :paramref:`~duplicate.seperator`.

        separator: str, default='_'
           String used for adding the :paramref:`~duplicate.prefix` and the
           :paramref:`~duplicate.suffix` to every node's :attr:`Uid.name
           <tessif.frused.namedtuples.Uid>`.

        suffix: str, default=''
           String added to the beginning of every node's :attr:`Uid.name
           <tessif.frused.namedtuples.Uid>`, separated by
           :paramref:`~duplicate.seperator`.
        """
        duplicated_nodes = list()
        for node in self.nodes:
            duplicated_nodes.append(node.duplicate(
                prefix=prefix,
                separator=separator,
                suffix=suffix))

        return self.from_components(
            uid=self.uid,
            components=duplicated_nodes,
            timeframe=self.timeframe,
            global_constraints=self.global_constraints)

    @property
    def timeframe(self):
        return self._timeframe

    def pickle(self, location):
        """Pickle this system model.

        Parameters
        ----------
        location: str, default = None
            String representing of a path the created system model is dumped
            to. Passed to: meth: `pickle.dump`.
        """
        pickle.dump(self.__dict__, open(location, "wb"))

    def unpickle(self, location):
        """Restore a pickled energy system object."""
        self.__dict__ = pickle.load(open(location, "rb"))

    def tropp(self, what=None, py="3.8", tool=None, version=None):
        """TRansform, Optimize and PostProcess this Tessif system model."""
        what = "oemof_4.4_3.8"

        if not tool:
            tool = what.split(uid_seperator)[0]
        if not version:
            tool_version = what.split(uid_seperator)[1]
        if not py:
            py = what.split(uid_seperator)[2]

        requested_plugin = f"{tool}-{tool_version}"
        registered_plugin = registered_plugins[requested_plugin]
        registered_plugin += "-python"

        with tempfile.TemporaryDirectory() as tempdir:

            system_model_location = os.path.join(
                tempdir, "tessif_system_model.tsf")
            self.pickle(system_model_location)

            subprocess.run(
                [
                    "nox",
                    "-f",
                    # try absolutifying once fully ported
                    "/home/tze/Matze/Codes/tsf_release/test_nox.py",
                    "-R",
                    "-s",
                    registered_plugin,
                    "--python",
                    "3.8",
                    "--",
                    system_model_location,
                ]
            )

    @classmethod
    def from_components(cls, uid, components, timeframe,
                        global_constraints={'emissions': float('+inf')},
                        **kwargs):
        """
        Create an energy system from a collection of component instances.

        Particularly usefull when creating energy systems out of existing ones.

        Parameters
        ----------
        uid: ~collections.abc.Hashable
            Hashable unique identifier. Usually a string aka a name.

        components: `~collections.abc.Iterable`
            Iterable of :mod:`tessif.model.components.AbstractEsComponent`
            objects the energy system will be created of.

        timeframe: pandas.DatetimeIndex, optional
            Datetime index representing the evaluated timeframe. Explicitly
            stating:

                - initial datatime
                  (0th element of the :class:`pandas.DatetimeIndex`)
                - number of timesteps (length of :class:`pandas.DatetimeIndex`)
                - temporal resolution (:attr:`pandas.DatetimeIndex.freq`)

            For example::

                idx = pd.DatetimeIndex(
                    data=pd.date_range(
                        '2016-01-01 00:00:00', periods=11, freq='H'))

        global_constraints: dict, default={'emissions': float('+inf')}
            Dictionairy of :class:`numeric <numbers.Number>` values mapped to
            global constraint naming :class:`strings <str>`.

            Recognized constraint keys are:

                - ``emissions``
                - ...

            For a more detailed explanation see the user guide's section:
            :ref:`Secondary_Objectives`

        Return
        ------
        :class:`AbstractEnergySystem`
            The newly constructed energy system containing each component
            found in :paramref:`~from_components.components`.
        """
        busses, chps, sinks, sources, transformers = [], [], [], [], []
        connectors, storages = [], []

        for c in components:
            if isinstance(c, tessif_components.Bus):
                busses.append(c)
            elif isinstance(c, tessif_components.CHP):
                chps.append(c)
            elif isinstance(c, tessif_components.Sink):
                sinks.append(c)
            elif isinstance(c, tessif_components.Source):
                sources.append(c)
            elif isinstance(c, tessif_components.Transformer):
                transformers.append(c)
            elif isinstance(c, tessif_components.Connector):
                connectors.append(c)
            elif isinstance(c, tessif_components.Storage):
                storages.append(c)

        es = AbstractEnergySystem(
            uid=uid,
            busses=busses,
            chps=chps,
            sinks=sinks,
            sources=sources,
            transformers=transformers,
            connectors=connectors,
            storages=storages,
            timeframe=timeframe,
            global_constraints=global_constraints,
        )

        return es

    # def to_nxgrph(self):
    #     """
    #     Transform the :class:`AbstractEnergySystem` object into a
    #     :class:`networkx.DiGraph` object.

    #     Return
    #     ------
    #     directional_graph: networkx.DiGraph
    #         Networkx Directional Graph representing the abstract energy system.
    #     """
    #     grph = nx.DiGraph(name=self._uid)
    #     for node in self.nodes:
    #         grph.add_node(
    #             str(node.uid),
    #             **node.attributes,
    #         )

    #     nxgrph.create_edges(grph, self.edges, carrier=self._edge_carriers())
    #     return grph



# tessif/transform/mapping2es/tsf/__init__.py
"""
:mod:`~tessif.transform.mapping2es.tsf` is a :mod:`tessif` package to transform
abstract energy system data represented as `mapping
<https://docs.python.org/3/library/stdtypes.html#mapping-types-dict>`_ into
ready to simulate energy system model instances.

Mapping like representations are usually returned by utilities part of the
:mod:`~tessif.parse` module.
"""
from tessif.write import log
import pandas as pd
from tessif.frused import resolutions, spellings, configurations
from tessif.frused.defaults import energy_system_nodes as esn
from tessif.model import components, energy_system
import tessif.namedtuples as nts


def _generate_busses(energy_system_dict):
    """
    Generate :class:`~tessif.model.components.Bus` objects
    out of the :paramref:`~generate_busses.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionary of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_busses: ~collections.abc.Generator
        Generator object yielding the parsed
        :class:`bus objects <tessif.model.components.Bus>`.

    """
    busses = spellings.get_from(energy_system_dict, smth_like='bus',
                                dflt=pd.DataFrame())

    for row, bus in busses.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed
        if spellings.get_from(bus, smth_like='active', dflt=esn['active']):

            yield components.Bus(
                name=spellings.get_from(
                    bus, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    bus, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    bus, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    bus, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    bus, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    bus, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    bus, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    bus, smth_like='node_type',
                    dflt=esn['node_type']),
                inputs=spellings.get_from(
                    bus, smth_like='input',
                    dflt=esn['input']),
                outputs=spellings.get_from(
                    bus, smth_like='output',
                    dflt=esn['output']),)


[docs]
@log.timings
def _generate_sinks(energy_system_dict):
    """
    Generate :class:`~tessif.model.components.Sink` objects
    out of the :paramref:`~generate_sinks.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_sinks: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`sink objects <tessif.model.components.Sink>`.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Using the pure_python energy system mapping example

    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> sinks = ttsf._generate_sinks(parse.python_mapping(m))
    """
    sinks = spellings.get_from(energy_system_dict, smth_like='sink',
                               dflt=pd.DataFrame())

    for row, sink in sinks.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed
        if spellings.get_from(sink, smth_like='active', dflt=esn['active']):

            # parse inputs beforehand to set adequate defaults
            inputs = *spellings.get_from(
                sink, smth_like='inputs', dflt=esn['input']),

            yield components.Sink(
                name=spellings.get_from(
                    sink, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    sink, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    sink, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    sink, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    sink, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    sink, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    sink, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    sink, smth_like='node_type',
                    dflt=esn['node_type']),
                inputs=inputs,

                # linear problem parameters
                accumulated_amounts=spellings.get_from(
                    sink, smth_like='accumulated_amounts',
                    dflt={k: nts.MinMax(
                        min=esn['minimum'], max=esn['maximum'])
                        for k in inputs}),

                flow_rates=spellings.get_from(
                    sink, smth_like='flow_rates',
                    dflt={k: nts.MinMax(
                        esn['minimum'], esn['maximum']) for k in inputs}),

                flow_costs=spellings.get_from(
                    sink, smth_like='flow_costs',
                    dflt={k: esn['flow_costs'] for k in inputs}),

                flow_emissions=spellings.get_from(
                    sink, smth_like='flow_emissions',
                    dflt={k: esn['emissions'] for k in inputs}),

                flow_gradients=spellings.get_from(
                    sink, smth_like='flow_gradients',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient'], esn['negative_gradient'])
                        for k in inputs}),

                gradient_costs=spellings.get_from(
                    sink, smth_like='gradient_costs',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient_costs'],
                        esn['negative_gradient_costs'])
                        for k in inputs}),

                timeseries=spellings.get_from(
                    sink, smth_like='timeseries',
                    dflt=esn['timeseries']),

                # expansion problem parameters
                expandable=spellings.get_from(
                    sink, smth_like='expandable',
                    dflt={k: esn['expandable'] for k in inputs}),

                expansion_costs=spellings.get_from(
                    sink, smth_like='expansion_costs',
                    dflt={k: esn['expansion_costs'] for k in inputs}),

                expansion_limits=spellings.get_from(
                    sink, smth_like='expansion_limits',
                    dflt={k: nts.MinMax(
                        esn['minimum_expansion'],
                        esn['maximum_expansion']) for k in inputs}),

                # mixed integer linear problem
                milp=spellings.get_from(
                    sink, smth_like='milp',
                    dflt={k: esn['milp'] for k in inputs}),

                initial_status=spellings.get_from(
                    sink, smth_like='initial_status',
                    dflt=esn['initial_status']),

                status_inertia=spellings.get_from(
                    sink, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['minimum_uptime'],
                        esn['minimum_downtime'])),

                status_changing_costs=spellings.get_from(
                    sink, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['startup_costs'],
                        esn['shutdown_costs'])),

                number_of_status_changes=spellings.get_from(
                    sink, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['maximum_shutdowns'],
                        esn['maximum_startups'])),

                costs_for_being_active=spellings.get_from(
                    sink, smth_like='costs_for_being_active',
                    dflt=esn['costs_for_being_active']),
            )


[docs]
@log.timings
def _generate_sources(energy_system_dict):
    """
    Generate :class:`~tessif.model.components.Source`
    objects out of the :paramref:`~generate_sources.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_sources: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`source<tessif.model.components.Source>`
        objects.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Using the pure_python energy system mapping example

    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> sources = ttsf._generate_sources(parse.python_mapping(m)
    """
    sources = spellings.get_from(energy_system_dict, smth_like='source',
                                 dflt=pd.DataFrame())

    for row, source in sources.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed
        if spellings.get_from(source, smth_like='active', dflt=esn['active']):

            # parse outputs beforehand to set adequate defaults
            outputs = *spellings.get_from(
                source, smth_like='outputs', dflt=esn['output']),

            yield components.Source(
                name=spellings.get_from(
                    source, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    source, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    source, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    source, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    source, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    source, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    source, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    source, smth_like='node_type',
                    dflt=esn['node_type']),
                outputs=outputs,

                # linear problem parameters
                accumulated_amounts=spellings.get_from(
                    source, smth_like='accumulated_amounts',
                    dflt={k: nts.MinMax(
                        min=esn['minimum'], max=esn['maximum'])
                        for k in outputs}),

                flow_rates=spellings.get_from(
                    source, smth_like='flow_rates',
                    dflt={k: nts.MinMax(
                        esn['minimum'], esn['maximum']) for k in outputs}),

                flow_costs=spellings.get_from(
                    source, smth_like='flow_costs',
                    dflt={k: esn['flow_costs'] for k in outputs}),

                flow_emissions=spellings.get_from(
                    source, smth_like='flow_emissions',
                    dflt={k: esn['emissions'] for k in outputs}),

                flow_gradients=spellings.get_from(
                    source, smth_like='flow_gradients',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient'], esn['negative_gradient'])
                        for k in outputs}),

                gradient_costs=spellings.get_from(
                    source, smth_like='gradient_costs',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient_costs'],
                        esn['negative_gradient_costs'])
                        for k in outputs}),

                timeseries=spellings.get_from(
                    source, smth_like='timeseries',
                    dflt=esn['timeseries']),

                # expansion problem parameters
                expandable=spellings.get_from(
                    source, smth_like='expandable',
                    dflt={k: esn['expandable'] for k in outputs}),

                expansion_costs=spellings.get_from(
                    source, smth_like='expansion_costs',
                    dflt={k: esn['expansion_costs'] for k in outputs}),

                expansion_limits=spellings.get_from(
                    source, smth_like='expansion_limits',
                    dflt={k: nts.MinMax(
                        esn['minimum_expansion'],
                        esn['maximum_expansion']) for k in outputs}),

                # mixed integer linear problem
                milp=spellings.get_from(
                    source, smth_like='milp',
                    dflt={k: esn['milp'] for k in outputs}),

                initial_status=spellings.get_from(
                    source, smth_like='initial_status',
                    dflt=esn['initial_status']),

                status_inertia=spellings.get_from(
                    source, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['minimum_uptime'],
                        esn['minimum_downtime'])),

                status_changing_costs=spellings.get_from(
                    source, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['startup_costs'],
                        esn['shutdown_costs'])),

                number_of_status_changes=spellings.get_from(
                    source, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['maximum_shutdowns'],
                        esn['maximum_startups'])),

                costs_for_being_active=spellings.get_from(
                    source, smth_like='costs_for_being_active',
                    dflt=esn['costs_for_being_active']),
            )


[docs]
@log.timings
def _generate_transformers(energy_system_dict):
    r"""
    Generate :class:`~tessif.model.components.Transformer`
    objects out of the :paramref:`~generate_transformers.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_transformers: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`transformer<tessif.model.components.Transformer>`
        objects.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Using the pure_python energy system mapping example

    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> transformers = ttsf._generate_transformers(parse.python_mapping(m))
    """
    transformers = spellings.get_from(
        energy_system_dict, smth_like='transformer',
        dflt=pd.DataFrame())

    for row, transformer in transformers.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed

        if spellings.get_from(
                transformer, smth_like='active', dflt=esn['active']):

            # parse inputs beforehand to set adequate defaults
            inputs = *spellings.get_from(
                transformer, smth_like='inputs', dflt=esn['input']),

            # parse outputs beforehand to set adequate defaults
            outputs = *spellings.get_from(
                transformer, smth_like='outputs', dflt=esn['output']),

            yield components.Transformer(
                name=spellings.get_from(
                    transformer, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    transformer, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    transformer, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    transformer, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    transformer, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    transformer, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    transformer, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    transformer, smth_like='node_type',
                    dflt=esn['node_type']),

                inputs=inputs,
                outputs=outputs,
                conversions=spellings.get_from(
                    transformer, smth_like='efficiency',
                    dflt=esn['efficiency']),

                # linear problem parameters
                flow_rates=spellings.get_from(
                    transformer, smth_like='flow_rates',
                    dflt={k: nts.MinMax(
                        esn['minimum'], esn['maximum'])
                        for k in [*inputs, *outputs]}),

                flow_costs=spellings.get_from(
                    transformer, smth_like='flow_costs',
                    dflt={k: esn['flow_costs'] for k in [*inputs, *outputs]}),

                flow_emissions=spellings.get_from(
                    transformer, smth_like='flow_emissions',
                    dflt={k: esn['emissions'] for k in [*inputs, *outputs]}),

                flow_gradients=spellings.get_from(
                    transformer, smth_like='flow_gradients',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient'], esn['negative_gradient'])
                        for k in [*inputs, *outputs]}),

                gradient_costs=spellings.get_from(
                    transformer, smth_like='gradient_costs',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient_costs'],
                        esn['negative_gradient_costs'])
                        for k in [*inputs, *outputs]}),

                timeseries=spellings.get_from(
                    transformer, smth_like='timeseries',
                    dflt=esn['timeseries']),

                # expansion problem parameters
                expandable=spellings.get_from(
                    transformer, smth_like='expandable',
                    dflt={k: esn['expandable'] for k in [*inputs, *outputs]}),

                expansion_costs=spellings.get_from(
                    transformer, smth_like='expansion_costs',
                    dflt={k: esn['expansion_costs']
                          for k in [*inputs, *outputs]}),

                expansion_limits=spellings.get_from(
                    transformer, smth_like='expansion_limits',
                    dflt={k: nts.MinMax(
                        esn['minimum_expansion'],
                        esn['maximum_expansion'])
                        for k in [*inputs, *outputs]}),

                # mixed integer linear problem
                milp=spellings.get_from(
                    transformer, smth_like='milp',
                    dflt={k: esn['milp']
                          for k in [*inputs, *outputs]}),

                initial_status=spellings.get_from(
                    transformer, smth_like='initial_status',
                    dflt=esn['initial_status']),

                status_inertia=spellings.get_from(
                    transformer, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['minimum_uptime'],
                        esn['minimum_downtime'])),

                status_changing_costs=spellings.get_from(
                    transformer, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['startup_costs'],
                        esn['shutdown_costs'])),

                number_of_status_changes=spellings.get_from(
                    transformer, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['maximum_shutdowns'],
                        esn['maximum_startups'])),

                costs_for_being_active=spellings.get_from(
                    transformer, smth_like='costs_for_being_active',
                    dflt=esn['costs_for_being_active']),
            )


def _generate_chps(energy_system_dict):
    r"""
    Generate :class:`~tessif.model.components.CHP` objects out of the
    :paramref:`~generate_chps.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_chps: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`chp <tessif.model.components.CHP>`
        objects.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Store the create_variable_chp example from the hardcoded examples into a
    hdf5 file and then parse it into a mapping from which the chps are
    generated.

    >>> import tessif.examples.data.tsf.py_hard as tsf_examples
    >>> tsf_es = tsf_examples.create_variable_chp()
    >>> import os
    >>> from tessif.frused.paths import write_dir
    >>> output_msg = tsf_es.to_hdf5(
    ...     directory=os.path.join(write_dir, 'tsf'),
    ...     filename='chp_es.hdf5',
    ... )
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> chps = ttsf._generate_chps(parse.hdf5(
    ...     path=os.path.join(write_dir, 'tsf', 'chp_es.hdf5'),))
    >>> for c in chps:
    ...     for name, atr in c.attributes.items():
    ...         print('{}: {}'.format(
    ...             name, sorted(atr) if isinstance(atr, frozenset) else atr))
    ... # frozensets are printed as sorted lists for doctesting consistency
    back_pressure: nan
    conversion_factor_full_condensation: {('gas', 'electricity'): 0.5}
    conversions: {('gas', 'electricity'): 0.3, ('gas', 'heat'): 0.2}
    costs_for_being_active: 0.0
    el_efficiency_wo_dist_heat: MinMax(min=nan, max=nan)
    enthalpy_loss: MinMax(min=nan, max=nan)
    expandable: {'electricity': False, 'gas': False, 'heat': False}
    expansion_costs: {'electricity': 0.0, 'gas': 0.0, 'heat': 0.0}
    expansion_limits: {'electricity': MinMax(min=0.0, max=inf), 'gas': MinMax(min=0.0, max=inf), 'heat': MinMax(min=0.0, max=inf)}
    flow_costs: {'electricity': 3, 'gas': 0, 'heat': 2}
    flow_emissions: {'electricity': 2, 'gas': 0, 'heat': 3}
    flow_gradients: {'electricity': PositiveNegative(positive=inf, negative=inf), 'gas': PositiveNegative(positive=inf, negative=inf), 'heat': PositiveNegative(positive=inf, negative=inf)}
    flow_rates: {'electricity': MinMax(min=0, max=9), 'gas': MinMax(min=0.0, max=inf), 'heat': MinMax(min=0, max=6)}
    gradient_costs: {'electricity': PositiveNegative(positive=0.0, negative=0.0), 'gas': PositiveNegative(positive=0.0, negative=0.0), 'heat': PositiveNegative(positive=0.0, negative=0.0)}
    initial_status: 1
    inputs: ['gas']
    interfaces: ['electricity', 'gas', 'heat']
    milp: {'electricity': False, 'gas': False, 'heat': False}
    min_condenser_load: nan
    number_of_status_changes: OnOff(on=0, off=0)
    outputs: ['electricity', 'heat']
    power_loss_index: nan
    power_wo_dist_heat: MinMax(min=nan, max=nan)
    status_changing_costs: OnOff(on=0, off=0)
    status_inertia: OnOff(on=0, off=0)
    timeseries: None
    uid: CHP1
    back_pressure: False
    conversion_factor_full_condensation: {}
    conversions: {}
    costs_for_being_active: 0.0
    el_efficiency_wo_dist_heat: MinMax(min=[0.43, 0.43, 0.43, 0.43], max=[0.53, 0.53, 0.53, 0.53])
    enthalpy_loss: MinMax(min=[1.0, 1.0, 1.0, 1.0], max=[0.18, 0.18, 0.18, 0.18])
    expandable: {'electricity': False, 'gas': False, 'heat': False}
    expansion_costs: {'electricity': 0.0, 'gas': 0.0, 'heat': 0.0}
    expansion_limits: {'electricity': MinMax(min=0.0, max=inf), 'gas': MinMax(min=0.0, max=inf), 'heat': MinMax(min=0.0, max=inf)}
    flow_costs: {'electricity': 3, 'gas': 0, 'heat': 2}
    flow_emissions: {'electricity': 2, 'gas': 0, 'heat': 3}
    flow_gradients: {'electricity': PositiveNegative(positive=inf, negative=inf), 'gas': PositiveNegative(positive=inf, negative=inf), 'heat': PositiveNegative(positive=inf, negative=inf)}
    flow_rates: {'electricity': MinMax(min=0.0, max=inf), 'gas': MinMax(min=0.0, max=inf), 'heat': MinMax(min=0.0, max=inf)}
    gradient_costs: {'electricity': PositiveNegative(positive=0.0, negative=0.0), 'gas': PositiveNegative(positive=0.0, negative=0.0), 'heat': PositiveNegative(positive=0.0, negative=0.0)}
    initial_status: 1
    inputs: ['gas']
    interfaces: ['electricity', 'gas', 'heat']
    milp: {'electricity': False, 'gas': False, 'heat': False}
    min_condenser_load: [3, 3, 3, 3]
    number_of_status_changes: OnOff(on=0, off=0)
    outputs: ['electricity', 'heat']
    power_loss_index: [0.19, 0.19, 0.19, 0.19]
    power_wo_dist_heat: MinMax(min=[8, 8, 8, 8], max=[20, 20, 20, 20])
    status_changing_costs: OnOff(on=0, off=0)
    status_inertia: OnOff(on=0, off=0)
    timeseries: None
    uid: CHP2
    """
    chps = spellings.get_from(
        energy_system_dict, smth_like='combined_heat_power',
        dflt=pd.DataFrame())

    for row, chp in chps.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed

        if spellings.get_from(
                chp, smth_like='active', dflt=esn['active']):
            # parse inputs beforehand to set adequate defaults
            inputs = *spellings.get_from(
                chp, smth_like='inputs', dflt=esn['input']),

            # parse outputs beforehand to set adequate defaults
            outputs = *spellings.get_from(
                chp, smth_like='outputs', dflt=esn['output']),

            yield components.CHP(
                name=spellings.get_from(
                    chp, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    chp, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    chp, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    chp, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    chp, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    chp, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    chp, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    chp, smth_like='node_type',
                    dflt=esn['node_type']),

                inputs=inputs,
                outputs=outputs,
                back_pressure=spellings.get_from(
                    chp, smth_like='back_pressure',
                    dflt=esn['chp_back_pressure']),

                conversion_factor_full_condensation=spellings.get_from(
                    chp, smth_like='conversion_factor_full_condensation',
                    dflt=esn['chp_efficiency']),

                el_efficiency_wo_dist_heat=spellings.get_from(
                    chp, smth_like='el_efficiency_wo_dist_heat',
                    dflt=esn['el_efficiency_wo_dist_heat']),

                enthalpy_loss=spellings.get_from(
                    chp, smth_like='enthalpy_loss',
                    dflt=esn['enthalpy_loss']),

                min_condenser_load=spellings.get_from(
                    chp, smth_like='min_condenser_load',
                    dflt=esn['min_condenser_load']),

                power_loss_index=spellings.get_from(
                    chp, smth_like='power_loss_index',
                    dflt=esn['power_loss_index']),

                power_wo_dist_heat=spellings.get_from(
                    chp, smth_like='power_wo_dist_heat',
                    dflt=esn['power_wo_dist_heat']),

                conversions=spellings.get_from(
                    chp, smth_like='efficiency',
                    dflt=esn['chp_efficiency']),

                # linear problem parameters
                flow_rates=spellings.get_from(
                    chp, smth_like='flow_rates',
                    dflt={k: nts.MinMax(
                        esn['minimum'], esn['maximum'])
                        for k in [*inputs, *outputs]}),

                flow_costs=spellings.get_from(
                    chp, smth_like='flow_costs',
                    dflt={k: esn['flow_costs'] for k in [*inputs, *outputs]}),

                flow_emissions=spellings.get_from(
                    chp, smth_like='flow_emissions',
                    dflt={k: esn['emissions'] for k in [*inputs, *outputs]}),

                flow_gradients=spellings.get_from(
                    chp, smth_like='flow_gradients',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient'], esn['negative_gradient'])
                        for k in [*inputs, *outputs]}),

                gradient_costs=spellings.get_from(
                    chp, smth_like='gradient_costs',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient_costs'],
                        esn['negative_gradient_costs'])
                        for k in [*inputs, *outputs]}),

                timeseries=spellings.get_from(
                    chp, smth_like='timeseries',
                    dflt=esn['timeseries']),

                # expansion problem parameters
                expandable=spellings.get_from(
                    chp, smth_like='expandable',
                    dflt={k: esn['expandable'] for k in [*inputs, *outputs]}),

                expansion_costs=spellings.get_from(
                    chp, smth_like='expansion_costs',
                    dflt={k: esn['expansion_costs']
                          for k in [*inputs, *outputs]}),

                expansion_limits=spellings.get_from(
                    chp, smth_like='expansion_limits',
                    dflt={k: nts.MinMax(
                        esn['minimum_expansion'],
                        esn['maximum_expansion'])
                        for k in [*inputs, *outputs]}),

                # mixed integer linear problem
                milp=spellings.get_from(
                    chp, smth_like='milp',
                    dflt={k: esn['milp']
                          for k in [*inputs, *outputs]}),

                initial_status=spellings.get_from(
                    chp, smth_like='initial_status',
                    dflt=esn['initial_status']),

                status_inertia=spellings.get_from(
                    chp, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['minimum_uptime'],
                        esn['minimum_downtime'])),

                status_changing_costs=spellings.get_from(
                    chp, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['startup_costs'],
                        esn['shutdown_costs'])),

                number_of_status_changes=spellings.get_from(
                    chp, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['maximum_shutdowns'],
                        esn['maximum_startups'])),

                costs_for_being_active=spellings.get_from(
                    chp, smth_like='costs_for_being_active',
                    dflt=esn['costs_for_being_active']),
            )


[docs]
@log.timings
def _generate_storages(energy_system_dict):
    """
    Generate :class:`~tessif.model.components.Storage`
    objects out of the :paramref:`~generate_storages.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_storages: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`storage<tessif.model.components.Storage>`
        objects.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Using the pure_python energy system mapping example

    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> storages = ttsf._generate_storages(parse.python_mapping(m))
    >>> for s in storages:
    ...     print(s.uid)
    ...     print(s.input)
    ...     print(s.output)
    ...     for name, atr in s.attributes.items():
    ...         print('{}: {}'.format(
    ...             name, sorted(atr) if isinstance(atr, frozenset) else atr))
    ...     # frozensets are printed as sorted lists for doctesting consistency
    Battery
    electricity
    electricity
    capacity: 100
    costs_for_being_active: 0
    expandable: {'electricity': False}
    expansion_costs: {'electricity': 0}
    expansion_limits: {'electricity': MinMax(min=0, max=inf)}
    final_soc: None
    fixed_expansion_ratios: {'electricity': True}
    flow_costs: {'electricity': 0}
    flow_efficiencies: {'electricity': InOut(inflow=0.95, outflow=0.98)}
    flow_emissions: {'electricity': 0}
    flow_gradients: {'electricity': PositiveNegative(positive=10, negative=10)}
    flow_rates: {'electricity': MinMax(min=0, max=10)}
    gradient_costs: {'electricity': PositiveNegative(positive=0, negative=0)}
    idle_changes: PositiveNegative(positive=0, negative=1)
    initial_soc: 10
    initial_status: True
    input: electricity
    interfaces: ['electricity']
    milp: {'electricity': False}
    number_of_status_changes: OnOff(on=0, off=2)
    output: electricity
    status_changing_costs: OnOff(on=0, off=2)
    status_inertia: OnOff(on=0, off=2)
    timeseries: None
    uid: Battery
    """
    storages = spellings.get_from(
        energy_system_dict, smth_like='generic_storage',
        dflt=pd.DataFrame())

    for row, storage in storages.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed

        if spellings.get_from(
                storage, smth_like='active', dflt=esn['active']):

            # parse inputs beforehand to set adequate defaults
            input = spellings.get_from(
                storage, smth_like='input', dflt=esn['input'])

            # parse outputs beforehand to set adequate defaults
            output = spellings.get_from(
                storage, smth_like='output', dflt=esn['output'])

            storage = components.Storage(
                name=spellings.get_from(
                    storage, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    storage, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    storage, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    storage, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    storage, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    storage, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    storage, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    storage, smth_like='node_type',
                    dflt=esn['node_type']),

                input=input,
                output=output,

                capacity=spellings.get_from(
                    storage, smth_like='storage_capacity',
                    dflt=esn['storage_capacity']),

                initial_soc=spellings.get_from(
                    storage, smth_like='initial_soc',
                    dflt=esn['initial_soc']),

                idle_changes=spellings.get_from(
                    storage, smth_like='idle_changes',
                    dflt=nts.PositiveNegative(
                        esn['gain_rate'], esn['loss_rate'])),

                flow_rates=spellings.get_from(
                    storage, smth_like='flow_rates',
                    dflt={k: nts.MinMax(
                        esn['minimum'], esn['maximum'])
                        for k in [input, output]}),

                flow_costs=spellings.get_from(
                    storage, smth_like='flow_costs',
                    dflt={k: esn['flow_costs'] for k in [input, output]}),

                flow_efficiencies=spellings.get_from(
                    storage, smth_like='flow_efficiencies',
                    dflt={k: nts.InOut(
                        inflow=esn['efficiency'], outflow=esn['efficiency'])
                        for k in [input, output]},
                ),

                flow_emissions=spellings.get_from(
                    storage, smth_like='flow_emissions',
                    dflt={k: esn['emissions'] for k in [input, output]}),

                flow_gradients=spellings.get_from(
                    storage, smth_like='flow_gradients',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient'], esn['negative_gradient'])
                        for k in [input, output]}),

                gradient_costs=spellings.get_from(
                    storage, smth_like='gradient_costs',
                    dflt={k: nts.PositiveNegative(
                        esn['positive_gradient_costs'],
                        esn['negative_gradient_costs'])
                        for k in [input, output]}),

                timeseries=spellings.get_from(
                    storage, smth_like='timeseries',
                    dflt=esn['timeseries']),

                # expansion problem parameters
                expandable=spellings.get_from(
                    storage, smth_like='expandable',
                    dflt={k: esn['expandable']
                          for k in [input, output, 'capacity']}),

                fixed_expansion_ratios=spellings.get_from(
                    storage, smth_like='fixed_expansion_ratios',
                    dflt={k: esn['fixed_expansion_ratios']
                          for k in [input, output]}),

                expansion_costs=spellings.get_from(
                    storage, smth_like='expansion_costs',
                    dflt={k: esn['expansion_costs']
                          for k in [input, output, 'capacity']}),

                expansion_limits=spellings.get_from(
                    storage, smth_like='expansion_limits',
                    dflt={k: nts.MinMax(
                        esn['minimum_expansion'],
                        esn['maximum_expansion'])
                        for k in [input, output, 'capacity']}),

                # mixed integer linear problem
                milp=spellings.get_from(
                    storage, smth_like='milp',
                    dflt={k: esn['milp']
                          for k in [input, output]}),

                initial_status=spellings.get_from(
                    storage, smth_like='initial_status',
                    dflt=esn['initial_status']),

                status_inertia=spellings.get_from(
                    storage, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['minimum_uptime'],
                        esn['minimum_downtime'])),

                status_changing_costs=spellings.get_from(
                    storage, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['startup_costs'],
                        esn['shutdown_costs'])),

                number_of_status_changes=spellings.get_from(
                    storage, smth_like='status_inertia',
                    dflt=nts.OnOff(
                        esn['maximum_shutdowns'],
                        esn['maximum_startups'])),

                costs_for_being_active=spellings.get_from(
                    storage, smth_like='costs_for_being_active',
                    dflt=esn['costs_for_being_active']),
            )

            yield storage


[docs]
@log.timings
def _generate_connectors(energy_system_dict):
    r"""
    Generate :class:`~tessif.model.components.Connector`
    objects out of the :paramref:`~generate_transformers.energy_system_dict`.

    Parameters
    ----------
    energy_system_dict: dict
        Dictionairy of :class:`pandas.DataFrame` objects keyed by strings.
        With the key strings representing energy system components.
        Usually returned by something like the :mod:`tessif.parse`
        functionalities.

    Return
    ------
    generated_connectors: ~collections.abc.Generator
        Generator obeject yielding the parsed
        :class:`connector <tessif.model.components.Connector>`
        objects.

    Examples
    --------
    Setting :attr:`spellings.get_from's <tessif.frused.spellings.get_from>`
    logging level to debug for decluttering doctest output:

    >>> from tessif.frused import configurations
    >>> configurations.spellings_logging_level = 'debug'

    Using the pure_python energy system mapping example

    >>> import tessif.examples.data.tsf.py_mapping.transshipment as tship
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> connectors = ttsf._generate_connectors(
    ...     parse.python_mapping(tship.mapping))
    >>> for c in connectors:
    ...     for name, atr in c.attributes.items():
    ...         print('{}: {}'.format(
    ...             name, sorted(atr) if isinstance(atr, frozenset) else atr))
    ... # frozensets are printed as sorted lists for doctesting consistency
    conversions: {('bus-01', 'bus-02'): 0.9, ('bus-02', 'bus-01'): 0.8}
    inputs: ['bus-01', 'bus-02']
    interfaces: ['bus-01', 'bus-02']
    outputs: ['bus-01', 'bus-02']
    timeseries: None
    uid: connector
    """
    connectors = spellings.get_from(
        energy_system_dict, smth_like='connector',
        dflt=pd.DataFrame())

    for row, connector in connectors.iterrows():
        # check active switch, but assume user forgot adding one
        # and meant to add all entries listed

        if spellings.get_from(
                connector, smth_like='active', dflt=esn['active']):

            # parse interface beforehand to set adequate defaults
            # a connectors interface is both it's inputs as well it's outputs
            interfaces = *spellings.get_from(
                connector, smth_like='interfaces', dflt=esn['interfaces']),

            yield components.Connector(
                name=spellings.get_from(
                    connector, smth_like='name',
                    dflt=esn['name']),
                latitude=spellings.get_from(
                    connector, smth_like='latitude',
                    dflt=esn['latitude']),
                longitude=spellings.get_from(
                    connector, smth_like='longitude',
                    dflt=esn['longitude']),
                region=spellings.get_from(
                    connector, smth_like='region',
                    dflt=esn['region']),
                sector=spellings.get_from(
                    connector, smth_like='sector',
                    dflt=esn['sector']),
                carrier=spellings.get_from(
                    connector, smth_like='carrier',
                    dflt=esn['carrier']),
                component=spellings.get_from(
                    connector, smth_like='component',
                    dflt=esn['component']),
                node_type=spellings.get_from(
                    connector, smth_like='node_type',
                    dflt=esn['node_type']),

                interfaces=interfaces,

                conversions=spellings.get_from(
                    connector, smth_like='efficiency',
                    dflt=esn['efficiency']),

                # # expansion problem parameters
                # expandable=spellings.get_from(
                #     connector, smth_like='expandable',
                #     dflt={k: esn['expandable'] for k in [*inputs, *outputs]}),

                # expansion_costs=spellings.get_from(
                #     connector, smth_like='expansion_costs',
                #     dflt={k: esn['expansion_costs']
                #           for k in [*inputs, *outputs]}),

                # expansion_limits=spellings.get_from(
                #     connector, smth_like='expansion_limits',
                #     dflt={k: nts.MinMax(
                #         esn['minimum_expansion'],
                #         esn['maximum_expansion'])
                #         for k in [*inputs, *outputs]}),

            )


[docs]
@log.timings
def infer_timeindex(energy_system_dict):
    """
    Try inferring the timeindex from the energy system dictionary.
    Inferring is done by reading out the
    :attr:`~tessif.frused.spellings.timeindex`-like column
    of the :attr:`~tessif.frused.spellings.timeseries`-like spreadsheet
    (:class:`~pandas.DataFrame`) and rounding it according to the
    :attr:`temporal resolution
    <tessif.frused.configurations.temporal_resolution>`.

    The timeframe to be simulated is here referred to as "timeindex".

    Warning
    -------
    For inferring to be succesfull, discrete timeframe length has to be at
    least 3 (after rounding).

    Parameters
    ----------
    energy_system_dict: dict
        Dictionary containing the energy system data. Typically returned by
        one of the :mod:`tessif.parse` functionalities

    Return
    ------
    timeindex: pandas.DatetimeIndex
        The time frame inferred from the energy system dictionary.

    Example
    -------
    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> from tessif.frused.paths import example_dir
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> idx = ttsf.infer_timeindex(parse.python_mapping(m))
    >>> import pprint
    >>> pprint.pprint(idx)
    DatetimeIndex(['1990-07-13 00:00:00', '1990-07-13 01:00:00',
                   '1990-07-13 02:00:00'],
                  dtype='datetime64[ns]', freq='H')
    """

    # find out which expression of 'Timeframe' was used
    ts_key = [variant for variant in spellings.timeframe
              if variant in energy_system_dict.keys()][0]

    # find out which expression of 'Timeindex' was used
    idx_key = [variant for variant in spellings.timeindex
               if variant in energy_system_dict[ts_key].columns][0]

    # logger.debug(
    #    ("Inferring timeframe using '{}' for timeframe and "
    #     "'{}' for timeindex".format(
    #        ts_key, idx_key))

    # Extract the time index
    timeindex = pd.DatetimeIndex(
        energy_system_dict[ts_key][idx_key].values, freq='infer')

    # Round to time index according to temporal resolution settings to...
    # correct rounding errors of the input source due to time format display
    timeindex = timeindex.round(
        resolutions.temporal_rounding_map[configurations.temporal_resolution])

    # Reinfer frequency from newly round time index
    timeindex = pd.DatetimeIndex(timeindex, freq='infer')

    # If reinferring was unsuccessful, enforce tessif's current frequency
    if timeindex.freq is None:
        timeindex.freq = resolutions.temporal_rounding_map[
            configurations.temporal_resolution]

    return timeindex


[docs]
@log.timings
def transform(energy_system_mapping, **kwargs):
    """
    Transform an :meth:`energy system mapping
    <tessif.parse.python_mapping>` to a :class:`tessif energy system
    <tessif.model.energy_system>`.

    Parameters
    ----------
    energy_system_mapping: dict
        Dictionairy containing the energy system data. Typically returned by
        one of the :mod:`tessif.parse` functionalities

    timeframe: pandas.DatetimeIndex
        Datetime index representing the evaluated timeframe. Explicitly
        stating:

            - initial datatime
              (0th element of the :class:`pandas.DatetimeIndex`)
            - number of timesteps (length of :class:`pandas.DatetimeIndex`)
            - temporal resolution (:attr:`pandas.DatetimeIndex.freq`)

        For example::

            idx = pd.DatetimeIndex(
                data=pd.date_range(
                    '2016-01-01 00:00:00', periods=11, freq='H'))

        Note
        ----
        If not stated it is tried to be :meth:`inferred<infer_timeindex>`.

    global_constraints: dict, default={'emissions': float('+inf')}
        Dictionairy of :class:`numeric <numbers.Number>` values mapped to
        global constraint naming :class:`strings <str>`.

        Recognized constraint keys are:

            - ``emissions``
            - ...

        For a more detailed explanation see the user guide's section:
        :ref:`Secondary_Objectives`

    uid: ~collections.abc.Hashable, default='automatically_transformed'
        Hashable unique identifier. Usually a string aka a name.

    Return
    ------
    energy_system: ~tessif.model.energy_system.AbstractEnergySystem
        The dictionairy transformed oemof energy system, ready for simulation.

    Examples
    --------
    Using the pure_python energy system mapping example:

    >>> from tessif.examples.data.tsf.py_mapping.fpwe import mapping as m
    >>> import tessif.parse as parse
    >>> import tessif.transform.mapping2es.tsf as ttsf
    >>> esys = ttsf.transform(parse.python_mapping(m))
    >>> for node in esys.nodes:
    ...     for k, v in node.attributes.items():
    ...         print('{} = {}'.format(
    ...             k, sorted(v) if isinstance(v, frozenset) else v))
    ... # frozensets are printed as sorted lists for doctesting consistency
    ...     print(50*'-')
    ...     print()
    inputs = ['Gas Station.fuel']
    interfaces = ['Gas Station.fuel', 'Generator.fuel']
    outputs = ['Generator.fuel']
    timeseries = None
    uid = Pipeline
    --------------------------------------------------
    <BLANKLINE>
    inputs = ['Battery.electricity', 'Generator.electricity']
    interfaces = ['Battery.electricity', 'Demand.electricity', 'Generator.electricity']
    outputs = ['Battery.electricity', 'Demand.electricity']
    timeseries = None
    uid = Power Line
    --------------------------------------------------
    <BLANKLINE>
    accumulated_amounts = {'fuel': MinMax(min=0, max=inf)}
    costs_for_being_active = 0
    expandable = {'fuel': False}
    expansion_costs = {'fuel': 0}
    expansion_limits = {'fuel': MinMax(min=0, max=inf)}
    flow_costs = {'fuel': 0}
    flow_emissions = {'fuel': 0}
    flow_gradients = {'fuel': PositiveNegative(positive=42, negative=42)}
    flow_rates = {'fuel': MinMax(min=0, max=22)}
    gradient_costs = {'fuel': PositiveNegative(positive=1, negative=1)}
    initial_status = True
    interfaces = ['fuel']
    milp = {'fuel': False}
    number_of_status_changes = OnOff(on=1, off=1)
    outputs = ['fuel']
    status_changing_costs = OnOff(on=1, off=1)
    status_inertia = OnOff(on=1, off=1)
    timeseries = {'fuel': MinMax(min=0, max=array([10, 22, 22]))}
    uid = Gas Station
    --------------------------------------------------
    <BLANKLINE>
    accumulated_amounts = {'electricity': MinMax(min=0, max=inf)}
    costs_for_being_active = 0
    expandable = {'electricity': False}
    expansion_costs = {'electricity': 0}
    expansion_limits = {'electricity': MinMax(min=0, max=inf)}
    flow_costs = {'electricity': 0}
    flow_emissions = {'electricity': 0}
    flow_gradients = {'electricity': PositiveNegative(positive=11, negative=11)}
    flow_rates = {'electricity': MinMax(min=8, max=11)}
    gradient_costs = {'electricity': PositiveNegative(positive=0, negative=0)}
    initial_status = True
    inputs = ['electricity']
    interfaces = ['electricity']
    milp = {'electricity': False}
    number_of_status_changes = OnOff(on=2, off=1)
    status_changing_costs = OnOff(on=2, off=1)
    status_inertia = OnOff(on=2, off=1)
    timeseries = None
    uid = Demand
    --------------------------------------------------
    <BLANKLINE>
    conversions = {('fuel', 'electricity'): 0.42}
    costs_for_being_active = 0
    expandable = {'electricity': False, 'fuel': False}
    expansion_costs = {'electricity': 0, 'fuel': 0}
    expansion_limits = {'electricity': MinMax(min=0, max=inf), 'fuel': MinMax(min=0, max=inf)}
    flow_costs = {'electricity': 0, 'fuel': 0}
    flow_emissions = {'electricity': 0, 'fuel': 0}
    flow_gradients = {'electricity': PositiveNegative(positive=10, negative=10), 'fuel': PositiveNegative(positive=50, negative=50)}
    flow_rates = {'electricity': MinMax(min=5, max=15), 'fuel': MinMax(min=0, max=50)}
    gradient_costs = {'electricity': PositiveNegative(positive=0, negative=0), 'fuel': PositiveNegative(positive=0, negative=0)}
    initial_status = True
    inputs = ['fuel']
    interfaces = ['electricity', 'fuel']
    milp = {'electricity': True, 'fuel': False}
    number_of_status_changes = OnOff(on=0, off=2)
    outputs = ['electricity']
    status_changing_costs = OnOff(on=0, off=2)
    status_inertia = OnOff(on=0, off=2)
    timeseries = {'electricity': MinMax(min=0, max=array([10, 22, 22]))}
    uid = Generator
    --------------------------------------------------
    <BLANKLINE>
    capacity = 100
    costs_for_being_active = 0
    expandable = {'electricity': False}
    expansion_costs = {'electricity': 0}
    expansion_limits = {'electricity': MinMax(min=0, max=inf)}
    final_soc = None
    fixed_expansion_ratios = {'electricity': True}
    flow_costs = {'electricity': 0}
    flow_efficiencies = {'electricity': InOut(inflow=0.95, outflow=0.98)}
    flow_emissions = {'electricity': 0}
    flow_gradients = {'electricity': PositiveNegative(positive=10, negative=10)}
    flow_rates = {'electricity': MinMax(min=0, max=10)}
    gradient_costs = {'electricity': PositiveNegative(positive=0, negative=0)}
    idle_changes = PositiveNegative(positive=0, negative=1)
    initial_soc = 10
    initial_status = True
    input = electricity
    interfaces = ['electricity']
    milp = {'electricity': False}
    number_of_status_changes = OnOff(on=0, off=2)
    output = electricity
    status_changing_costs = OnOff(on=0, off=2)
    status_inertia = OnOff(on=0, off=2)
    timeseries = None
    uid = Battery
    --------------------------------------------------
    <BLANKLINE>

    Get the inspected timeframe:

    >>> print(esys.timeframe)
    DatetimeIndex(['1990-07-13 00:00:00', '1990-07-13 01:00:00',
                   '1990-07-13 02:00:00'],
                  dtype='datetime64[ns]', freq='H')

    Use the automatically generated energy system to :meth:`transform
    <tessif.model.energy_system.Abstract EnergySystem.to_nxgrph>` it into a
    :class:`networkx.Graph` to facilitate plotting:

    >>> import tessif.visualize.nxgrph as nxv
    >>> import matplotlib.pyplot as plt
    >>> grph = esys.to_nxgrph()
    >>> drawings = nxv.draw_graph(
    ...     grph, node_color='green', edge_color='pink')
    >>> plt.draw()

    IGNORE:
    >>> title = plt.gca().set_title(
    ...     'transform.tsf module design example')
    >>> plt.pause(2)
    >>> plt.close('all')

    IGNORE

    .. image:: transformed_es_example.png
        :align: center
        :alt: alternate text
    """

    # get/default global constraints
    global_constraints = kwargs.pop(
        'global_constraints',
        energy_system_mapping.get(
            'global_constraints',
            {'emissions': float('+inf')}))

    # get/infer timeindex
    timeindex = kwargs.pop(
        'timeframe', infer_timeindex(energy_system_mapping))

    # get/autofill uid
    _uid = kwargs.pop(
        'uid', 'automatically_transformed')

    busses = _generate_busses(energy_system_mapping)
    sinks = _generate_sinks(energy_system_mapping)
    sources = _generate_sources(energy_system_mapping)
    transformers = _generate_transformers(energy_system_mapping)
    chps = _generate_chps(energy_system_mapping)
    storages = _generate_storages(energy_system_mapping)
    connectors = _generate_connectors(energy_system_mapping)

    # Create energy system object
    es = energy_system.AbstractEnergySystem(
        uid=_uid,
        busses=busses,
        connectors=connectors,
        sinks=sinks,
        sources=sources,
        transformers=transformers,
        chps=chps,
        storages=storages,
        timeframe=timeindex,
        global_constraints=global_constraints)

    return es



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tessif',
 'tessif.frused',
 'tessif.frused.hooks',
 'tessif.model',
 'tessif.transform.mapping2es']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'matplotlib>=3.6.3,<4.0.0',
 'nox-poetry>=1.0.2,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'strutils>=0.1.0,<0.2.0',
 'tessif-examples>=0.2.2,<0.3.0']

entry_points = \
{'console_scripts': ['tessif = tessif.cli:main_cli_entry']}

setup_kwargs = {
    'name': 'tessif',
    'version': '0.0.20',
    'description': 'Transforming Energy Supply System modell*I*ng Framework',
    'long_description': 'tessif (**T**\\ ransforming **E**\\ nergy **S**\\ upply **S**\\ ystem model\\ **I**\\ ng **F**\\ ramework)\n===================================================================================================\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/tessif.svg\n   :target: https://pypi.org/project/tessif/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/tessif.svg\n   :target: https://pypi.org/project/tessif/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/tessif\n   :target: https://pypi.org/project/tessif\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/tessif\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/tessif/latest.svg?label=Read%20the%20Docs\n   :target: https://tessif.readthedocs.io/\n   :alt: Read the documentation at https://tessif.readthedocs.io/\n.. |Tests| image:: https://github.com/tZ3ma/tessif/workflows/Tests/badge.svg\n   :target: https://github.com/tZ3ma/tessif/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/tZ3ma/tessif/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/tZ3ma/tessif\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nLatest Stable Version\n---------------------\nYou can install *Tessif* via pip_ from PyPI_.\nInstall using a console with your virtual environment activated:\n\n.. code-block:: console\n\n   $ pip install tessif\n\nLatest Development Version (potentially unstable)\n-------------------------------------------------\n\nYou can install the development version of *Tessif* via pip_ from TestPyPI_.\nInstall using a console with your virtual environment activated:\n\n.. code-block:: console\n\n   $ pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple/ tessif\n\nThis installs the TestPyPI_ version of tessif while resolving the dependencies on PyPI_.\n\nLicense\n-------\nDistributed under the terms of the `MIT license`_,\n*Tessif* is free and open source software.\n\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _TestPyPI: https://test.pypi.org/\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://tessif.readthedocs.io/en/latest/usage.html\n',
    'author': 'Mathias Ammon',
    'author_email': 'mathias.ammon@tuhh.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tZ3ma/tessif',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
